/**
 * XXX TODO
 *
 * Copyright 2014 Matteo Riondato <matteo@cs.brown.edu>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 */

#include <algorithm>
#include <cerrno>
#include <chrono>
#include <cmath>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <random>
#include <vector>

#include <unistd.h>

#include "stats.h"
#include "util.h"
#include "vc_sample.h"

using namespace std;
using namespace std::chrono;

bool verbose = false;

void usage(char *binary_name) {
	cerr << binary_name << ": static sampling algorithm to extract an (eps,delta)-approximation" << endl;
	cerr << "USAGE: " << binary_name << " [-d d_index] [-h] [-n] [-v] sample_size delta dataset" << endl;
	cerr << "Sample a number of transactions from dataset sufficient to get an (epsilon,delta)-approximation" << endl;
	cerr << "\t-d d_index: use the specified value as d-index" << endl;
	cerr << "\t-h : print this message and exit" << endl;
	cerr << "\t-n : do not create sample" << endl;
	cerr << "\t-v : verbose output" << endl;
}

int main(int argc, char **argv) {
	bool create_sample = true;
	unsigned int d_index = 0;
	int signed_d_index = 0;
	int opt;
	while ((opt = getopt(argc, argv, "d:hnv")) != -1) {
		switch (opt) {
		case 'd':
			signed_d_index = atoi(optarg);
			if (errno == ERANGE || signed_d_index <= 0) {
				cerr << ERROR_HEADER << "d_index must be positive" << endl;
				return EXIT_FAILURE;
			}
			d_index = (unsigned int) signed_d_index;
			break;
		case 'h':
			usage(argv[0]);
			return EXIT_SUCCESS;
			break;
		case 'n':
			create_sample = false;
			break;
		case 'v':
			verbose = true;
			break;
		}
	}
	if (optind != argc - 3) {
		cerr << ERROR_HEADER << "wrong number of arguments" << endl;
		return EXIT_FAILURE;
	} 
	int signed_sample_size = atoi(argv[argc-3]);
	if (errno == ERANGE || signed_sample_size <= 0) {
		cerr << ERROR_HEADER << "sample_size should be greater than 0" <<
			endl;
		return 1;
	}
	unsigned int sample_size = (unsigned int) signed_sample_size;
	double delta = strtod(argv[argc - 2], NULL);
	if (errno == ERANGE || delta >= 1.0 || delta <= 0.0) {
		cerr << ERROR_HEADER << 
			"delta should be greater than 0 and smaller than 1" 
			<< endl;
		return EXIT_FAILURE;
	}
	char *dataset = argv[argc - 1];
	double ds_size = get_size(dataset);

	// Initialize random number generator
#ifndef DEBUG
	unsigned seed = std::chrono::system_clock::now().time_since_epoch().count();
#else
	unsigned seed = 0;
#endif
	mt19937_64 generator(seed);
	uniform_int_distribution<int> distribution(0, ds_size - 1);

	auto start = steady_clock::now();

	if (d_index == 0) {
		d_index = get_d_index_bound(dataset);
	}
	auto d_index_end = steady_clock::now();
	double epsilon = get_epsilon(sample_size, delta, d_index);
	if (verbose) {
		cerr << VERBOSE_HEADER << "d_index: " << d_index << ", epsilon: " <<
			epsilon << ", time: " <<
			duration_cast<milliseconds>(d_index_end - start).count() << " ms" <<
			endl;
	}
	
	if (create_sample) {
		FILE *ds_FILE = fopen(dataset, "r");
		if (ds_FILE == NULL) {
			perror("Error opening input file");
			return errno;
		}

		// Add transactions to sample (printing them to stdout)
		vector<unsigned int> sampled_indices(sample_size, 0);
		for (unsigned int i = 0; i < sample_size; ++i) {
			sampled_indices[i] = distribution(generator);
		}
		sort(sampled_indices.begin(), sampled_indices.end());

		unsigned int transaction_index = -1;
		unsigned int sampled_index = 0;
		char line[TRANSACTION_LINE_MAXLEN];
		while (sampled_index < sample_size) {
			// Read line
			do {
				get_next_line(line, sizeof(line), ds_FILE);
				transaction_index++;
			} while (transaction_index < sampled_indices[sampled_index]);

			unsigned int base_transaction_id = sampled_indices[sampled_index];

			do {
				sampled_index++;
			} while (sampled_index < sample_size &&
					sampled_indices[sampled_index] ==
					base_transaction_id);
		}
	} // if (create_sample)

	auto end = steady_clock::now();
	auto elapsed = duration_cast<milliseconds>(end - start);
	auto d_index_elapsed = duration_cast<milliseconds>(d_index_end - start);


	if (verbose) {
		cerr << VERBOSE_HEADER << "sampling completed. Took " << elapsed.count()
			<< " ms (" << d_index_elapsed.count()
			<< " ms for d-index computation)." << endl;
	}
	
	// Print useful information to stderr
	cerr << "dataset,epsilon,delta,d_index,sample_size,sampling_time,d_index_time,runtime" << endl;
	cerr << (strrchr(dataset, '/') + 1) << COMMA << epsilon << COMMA <<
		delta << COMMA << d_index << COMMA << sample_size << COMMA <<
		elapsed.count() - d_index_elapsed.count() << COMMA <<
		d_index_elapsed.count() << COMMA << elapsed.count() << endl;

	return EXIT_SUCCESS;
}
